import functools

import numpy as np
from numpy.lib.stride_tricks import as_strided

import loupe


class fft2(loupe.core.Function):
    def __init__(self, input):
        self.input = loupe.asarray(input)
        super().__init__(self.input)

    def forward(self):
        return np.fft.fft2(self.input.data)

    def backward(self, grad):
        return np.fft.ifft2(grad)


class ifft2(loupe.core.Function):
    def __init__(self, input):
        self.input = loupe.asarray(input)
        super().__init__(self.input)

    def forward(self):
        return np.fft.ifft2(self.input.data)

    def backward(self, grad):
        return np.fft.fft2(grad)


class dft2(loupe.core.Function):
    """Compute the 2-dimensional discrete Fourier Transform.

    This function allows independent control over input shape, output shape,
    and output sampling by implementing the matrix triple product algorithm
    described in [1].

    Parameters
    ----------
    x : array_like
        Input array, can be complex. Must 
    alpha : (2,) array_like
        Output plane sampling frequency in terms of (row, col).
    shape : (2,) arrqy_like, optional
        Shape of the output. If `shape` is None (default), the shape of the input
        is used.
    shift : sequence of floats, optional
        Number of pixels in (r,c) to shift the DC pixel in the output plane with
        the origin centrally located in the plane. Default is (0, 0). Can be 
        2-dimensional to provide independent shifts for each Fourier transform
        if `x` is 3-dimensional.
    offset : sequence of floats, optional
        Offset of the center of the input `x` in terms of number of pixels in 
        (r,c). Default is (0, 0). Can be 2-dimensional to provide independent
        offsets for each Fourier transform if `x` is 3-dimensional.
    unitary : bool, optional
        Normalization flag. If True, a normalization is performed on the output
        such that the DFT operation is unitary. Default is True.
    axes : sequence of ints, optional
        Axes over which to compute the Fourier transform. If not given, the last 
        two axes are used.

    Returns
    -------
    out : :class:`~loupe.core.Function`
        The discrete Fourier transform of the input.

    Notes
    -----
    * `dft2` expects the DC pixel to be centered in the input array.
      The center is assumed to be at ``np.floor(n/2) + 1`` for length `n`.
      Note this is consistent with a shifted FFT performed as:
      ``F = ifftshift(fft2(fftshift(f)))``

    * Performing the DFT with ``alpha = 1/f.shape``, ``unitary = False``,
      and ``shape = f.shape`` is equivalent to a shifted FFT:

    .. code:: pycon

        >>> f = loupe.rand(size=(10, 10))
        >>> F = loupe.dft2(f, alpha=(1/10, 1/10), unitary=False, shape=f.shape)
        >>> G = np.fft.ifftshift(np.fft.fft2(np.fft.fftshift(f)))
        >>> np.allclose(F, G)
        True

    References
    ----------
    [1] Soummer, et. al. Fast computation of Lyot-style coronagraph propagation (2007)

    """
    def __init__(self, x, alpha, shape=None, shift=(0,0), offset=(0,0), unitary=True, axes=None):

        self.input = loupe.asarray(x)
        if self.input.ndim not in (2,3):
            raise ValueError("Input array must have ndim == 2 or 3")

        self.alpha = loupe.asarray(alpha)
        self.out_shape, self.axes = _cook_nd_args(self.input, shape, axes)
        self.iter_axis = _iter_axis(self.axes)

        # broadcast shift and offset as necessary
        depth = self.input.shape[self.iter_axis]

        self.shift = np.broadcast_to(shift, (depth, 2))
        self.offset = np.broadcast_to(offset, (depth, 2))
        self.unitary = unitary
        super().__init__(self.input, self.alpha)

    def forward(self):
        input = self.input.getdata()

        if input.ndim == 2:
            input = input[np.newaxis,:]

        alpha = np.broadcast_to(self.alpha.getdata(), (2,))

        # rearrange input so that it is arranged as
        # [iteration axis, rows, cols]
        input = np.moveaxis(input, [self.iter_axis, *self.axes], np.arange(3))
        
        result = []
        W_row, W_row_conj, W_col, W_col_conj = [], [], [], []

        for x, shift, offset in zip(input, self.shift, self.offset):
            Wr, Wc = _dft2_matrices(input.shape[1], input.shape[2],
                                    self.out_shape[0], self.out_shape[1],
                                    alpha[0], alpha[1],
                                    shift[0], shift[1],
                                    offset[0], offset[1])
            result.append(np.dot(Wr.dot(x), Wc))
            W_row.append(Wr)
            W_row_conj.append(np.conj(Wr))
            W_col.append(Wc)
            W_col_conj.append(np.conj(Wc))
        
        result = np.squeeze(result)

        if self.unitary:
            root_alpha = np.sqrt(np.abs(alpha))
            norm_coeff = np.prod(root_alpha)
        else:
            root_alpha = 1
            norm_coeff = 1

        # NOTE: result is purposely saved before we apply unitary normalization.
        # The result that is used in the backward function is expected to be
        # "clean" (i.e. unitary=False)
        self.cache_for_backward(input, alpha, W_row, W_row_conj, W_col, W_col_conj,
                                result, root_alpha, norm_coeff)

        if self.unitary:
            result *= norm_coeff

        return result

    def backward(self, grad):

        # unpack saved_inputs
        (input, alpha, W_row, W_row_conj, W_col, W_col_conj,
         result, root_alpha, norm_coeff) = self.cache

        # backward() needs the "clean" (non-unitary) gradient. If unitary = True
        # we need to multiply the incoming gradient by norm_coeff. Because
        # norm_coeff is set to 1 in forward() if unitary = False, it is safe to
        # always multiply grad by norm_coeff here.
        grad_clean = grad * norm_coeff

        if grad_clean.ndim == 2:
            grad_clean = grad_clean[np.newaxis,:]

        input_grad = []
        for n, x in enumerate(grad_clean):
            Wr, Wc = W_row_conj[n], W_col_conj[n]
            input_grad.append(np.dot(Wr.T.dot(x), Wc.T))
        input_grad = np.squeeze(input_grad)

        self.input.backward(input_grad)


def _cook_nd_args(a, s=None, axes=None):
    # slightly modified version of numpy's function of the
    # same name
    if s is None:
        if axes is None:
            if a.ndim == 2:
                s = list(a.shape)
            elif a.ndim == 3:
                s = list(a.shape[1:3])
            else:
                raise ValueError("Array must have ndim == 2 or 3")
        else:
            s = np.take(a.shape, axes)
    s = list(s)
    if axes is None:
        axes = list(range(-len(s), 0))
    if len(s) != len(axes):
        raise ValueError("Shape and axes have different lengths.")
    return s, axes


def _iter_axis(axes):
    mask = np.ones(3, dtype=bool)
    mask[axes] = False
    return np.squeeze(np.arange(3)[mask])


def _dft2_matrices(m, n, M, N, alphar, alphac, shiftr, shiftc, offsetr, offsetc):
    R, S, U, V = _dft2_coords(m, n, M, N)
    E1 = np.exp(-2.0 * 1j * np.pi * alphar * np.outer(R+offsetr, U-shiftr)).T
    E2 = np.exp(-2.0 * 1j * np.pi * alphac * np.outer(S+offsetc, V-shiftc))
    return E1, E2

@functools.lru_cache(maxsize=32)
def _dft2_coords(m, n, M, N):
    # R and S are (r,c) coordinates in the (m x n) input plane f
    # U and V are (r,c) coordinates in the (M x N) output plane F
    R = np.arange(m) - np.floor(m/2.0)
    S = np.arange(n) - np.floor(n/2.0)
    U = np.arange(M) - np.floor(M/2.0)
    V = np.arange(N) - np.floor(N/2.0)
    return R, S, U, V



import numpy as np
import loupe


class tensordot(loupe.core.Function):
    """Compute tensor dot product along specified axes.

    Parameters
    ----------
    a, b : array_like
        Tensors to dot.
    axes : int or (2,) array_like, optional
        * If an int N, sum over the last N axes of *a* and the first N axes of
          *b* in order. The sizes of the corresponding axes must match.
        * If array_like, sum over the corresponding axes of *a* in the first 
          entry and the corresponding axes of *b* in the second entry. Both
          elements in axes must have the same length.

    Returns
    -------
    out : :class:`~loupe.core.Function`
        The tensor dot product of the input.

    See also
    --------
    :class:`~loupe.einsum`

    """
    def __init__(self, a, b, axes=2):
        self.a = loupe.asarray(a)
        self.b = loupe.asarray(b)
        self.axes = np.asarray(axes)

        self.a_axes = None
        self.b_axes = None
        self.c_axes = None

        super().__init__(a,b)

    def forward(self):
        a = self.a.getdata()
        b = self.b.getdata()
        self.cache_for_backward(a, b)

        result, a_axes, b_axes, c_axes = _tensordot(a, b, self.axes, self.a_axes, self.b_axes, self.c_axes)

        self.a_axes = a_axes
        self.b_axes = b_axes
        self.c_axes = c_axes

        return result

    def backward(self, grad):
        a, b = self.cache

        if self.a.requires_grad:
            a_grad, *_ = _tensordot(grad, b, axes=self.axes, 
                                    a_axes=self.c_axes,
                                    b_axes=[self.b_axes[1], self.b_axes[0]],
                                    c_axes=self.a_axes)
            self.a.backward(a_grad)

        if self.b.requires_grad:
            b_grad, *_ = _tensordot(a, grad, axes=self.axes, 
                                    a_axes=[self.a_axes[1], self.a_axes[0]],
                                    b_axes=self.c_axes,
                                    c_axes=self.b_axes)
            self.b.backward(b_grad)


def _tensordot(a, b, axes, a_axes, b_axes, c_axes):

    # chainer.tensordot.forward()
    if a_axes is None or b_axes is None:
        a_axes = [[], []]  # 0:row axes, 1:col axes
        b_axes = [[], []]  # 0:row axes, 1:col axes
        if axes.ndim != 0:
            a_axes[1], b_axes[0] = axes
            if np.isscalar(a_axes[1]):
                a_axes[1] = a_axes[1],
            if np.isscalar(b_axes[0]):
                b_axes[0] = b_axes[0],
        else:
            a_axes[1] = range(a.ndim - axes, a.ndim)
            b_axes[0] = range(axes)
        a_range = range(a.ndim)
        a_axes[0] = [i for i in a_range if i not in a_axes[1]]
        b_range = range(b.ndim)
        b_axes[1] = [i for i in b_range if i not in b_axes[0]]
    
    # chainer.tensordot._tensordot()
    a_col_ndim = len(a_axes[1])
    b_row_ndim = len(b_axes[0])
    if a_col_ndim != b_row_ndim:
        raise ValueError('axes count mismatch')
    if a.ndim < a_col_ndim or b.ndim < b_row_ndim:
        raise ValueError('dimension of input arrays must be '
                         'greater equal to dot-axes count '
                         f'({a_col_ndim})')
    for a_axis, b_axis in zip(a_axes[1], b_axes[0]):
        if a.shape[a_axis] != b.shape[b_axis]:
            raise ValueError('shape mismatch')

    y = np.tensordot(a, b, axes=(tuple(a_axes[1]), tuple(b_axes[0])))

    if c_axes is not None:
        a_row_ndim = len(a_axes[0])
        b_col_ndim = len(b_axes[1])
        c_row_ndim = len(c_axes[0])
        c_col_ndim = len(c_axes[1])
        if a_row_ndim != c_row_ndim:
            raise ValueError('axes count mismatch')
        if b_col_ndim != c_col_ndim:
            raise ValueError('axes count mismatch')

        trans = [None for i in range(y.ndim)]
        table_a = [1 if i in a_axes[0] else 0 for i in range(a.ndim)]
        table_a = np.cumsum(table_a) - 1
        for i, c_axis in enumerate(c_axes[0]):
            trans[c_axis] = table_a[a_axes[0][i]]
        table_b = [1 if i in b_axes[1] else 0 for i in range(b.ndim)]
        table_b = np.cumsum(table_b) - 1
        for i, c_axis in enumerate(c_axes[1]):
            trans[c_axis] = table_b[b_axes[1][i]] + len(a_axes[0])
        for i, c_axis in enumerate(trans):
            if i != c_axis:
                y = np.transpose(y, trans)
                break
    
    # chainer.tensordot.forward()
    if c_axes is None:
        c_axes = [[], []]  # 0:row axes, 1:col axes
        c_row_ndim = len(a_axes[0])
        c_col_ndim = len(b_axes[1])
        c_axes[0] = range(c_row_ndim)
        c_axes[1] = range(c_row_ndim, c_row_ndim + c_col_ndim)

    return y, a_axes, b_axes, c_axes


__version__ = '1.3.1'

from loupe.core import array

from loupe.cost import sserror

from loupe.einsum import einsum

from loupe.fourier import dft2

from loupe.numeric import (
    add,
    subtract,
    multiply,
    power,
    exp,
    expc,
    sum,
    slice
)

from loupe.optimize import optimize

from loupe.shape import circle, circlemask, hexagon, slit

from loupe.special import absolute_square, abs_square, rebin, normavg

from loupe.tensordot import tensordot

from loupe.util import (
    asarray,
    zeros,
    zeros_like,
    ones,
    ones_like,
    rand,
    randn,
    centroid,
    shift,
    register,
    medfix2
)

from loupe.zernike import (
    zernike,
    zernike_compose,
    zernike_basis,
    zernike_fit,
    zernike_remove,
    zernike_coordinates
)



import numpy as np
import loupe

np.seterr(all='raise')

s = 'abcdefghijklmnopqrstuvwxyz'


class sserror(loupe.core.Function):
    r"""Compute the normalized sum squared error between two arrays.

    The normalized sum squared error between ideal data `f` and estimated data
    `g` is given by

    .. math::

        \mbox{error} = \frac{1}{N}\sum_n \frac{\sum{\left|g(n,x,y)-f(n,x,y)\right|^2}}{\sum{\left|f(n,x,y)\right|^2}}

    for :math:`n = 1, 2, ... N` independent measurements.

    If `gain_bias_invariant` is ``True``, the error metric is computed such 
    that the result is independent of relative scaling (gain) and offset 
    (bias) differences between `f` and  `g` [1].
    
    Parameters
    ----------
    model : array_like
        Estimated or modeled data
    data : array_like
        Ideal or measured data
    mask : array_like, optional
        Mask applied to the inputs where a True value indicates that the 
        corresponding element of the array is invalid. Mask must either have
        the same shape as the inputs or be broadcastable to the same shape and
        contain entries that are castable to bool. If None (default), the 
        inputs are not masked.
    gain_bias_invariant : bool, optional
        If True (default), the error is computed to be independent of any 
        gain or bias differences between the inputs. 

    Returns
    -------
    error : float
        Normalized root mean squared error

    References
    ----------
    [1] S. Thurman and J. Fienup, "Phase retrieval with signal bias", J. Opt. Soc. Am. A/Vol. 26, No. 4 (2009)
    
    [2] A. Jurling and J. Fienup, "Applications of algorithmic differentiation to phase retrieval algorithms", J. Opt. Soc. Am. A/Vol. 31, No. 7 (2014)
    
    """
    def __init__(self, model, data, mask=None, gain_bias_invariant=True):
        self.f = np.asarray(data)
        self.g = loupe.asarray(model)

        if not np.array_equal(self.f.shape, self.g.shape):
            raise ValueError(f'model and data shape mismatch: {self.g.shape} != {self.f.shape}')

        if mask is not None:
            if mask.shape == data.shape:
                self.mask = mask
            else:
                self.mask = np.broadcast_to(mask, self.f.shape)
        else:
            self.mask = mask

        self.gain_bias_invariant = gain_bias_invariant

        # _f, _K, and _D only depend on static input data f and the mask, so
        # we can precompute them here
        self._f = self._G(self.f, self.mask)
        
        if self.f.ndim == 3:
            self._K = self.f.shape[0]
        else:
            self._K = 1
        
        if self.mask is None:
            self._D = np.sum(self.f ** 2)
        else:
            self._D = np.sum(self.f ** 2 * self.mask)

        super().__init__(self.g)
        
    @staticmethod
    def _G(data, mask):
        # This function implements Eqs. 12 and 13 in [1]
        if mask is None:
            numer = np.sum(data, axis=(-2, -1))
            denom = np.prod(data.shape[-2:])
        else:
            numer = np.sum(mask * data, axis=(-2, -1))
            denom = np.sum(mask, axis=(-2, -1))
        x = numer/denom
        if data.ndim == 3:
            x = x[:, np.newaxis, np.newaxis]
        return data - x

    @staticmethod
    def _fit_gain(_f, _g, mask):
        # This function implements Eq. 14 in [1]
        try:
            if mask is None:
                numer = _f * _g
                denom = np.square(_g)
            else:
                numer = mask * _f * _g
                denom = mask * np.square(_g)
            
            return np.sum(numer, axis=(-2,-1))/np.sum(denom, axis=(-2,-1))

        except FloatingPointError as e:
            if np.all(_g == 0):
                raise ZeroDivisionError('g must be nonzero')
            else:
                raise e

    @staticmethod
    def _fit_bias(f, g, mask, gain):
        # This function implements Eq. 10 in [1]
        gain = np.asarray(gain)
        subscripts = f'{s[0:g.ndim]},{s[0:gain.ndim]}->{s[0:g.ndim]}'
        resid = f - np.einsum(subscripts, g, gain)
        if mask is None:
            numer = np.sum(resid, axis=(-2, -1))
            denom = np.prod(f.shape[-2:])
        else:
            numer = np.sum(mask*resid, axis=(-2, -2))
            denom = np.sum(mask, axis=(-2, -1))

        return numer/denom

    def residual(self):
        """Compute the residual error between `f` and `g`.

        Returns
        -------
        resid : ndarray
            Residual error between `f` and `g`.
        """
        g = self.g.getdata()

        if self.gain_bias_invariant:
            _g = self._G(g, self.mask)
            alpha = self._fit_gain(self._f, _g, self.mask)

            subscripts = f'{s[0:_g.ndim]},{s[0:alpha.ndim]}->{s[0:_g.ndim]}'

            if self.mask is None:
                resid = np.einsum(subscripts, _g, alpha) - self._f 
            else:
                resid = self.mask * (np.einsum(subscripts, _g, alpha) - self._f)
        else:
            if self.mask is None:
                resid = g - self.f
            else:
                resid = self.mask * (g - self.f)

        return resid

    def forward(self):
        g = self.g.getdata()

        if self.gain_bias_invariant:
            _g = self._G(g, self.mask)
            alpha = self._fit_gain(self._f, _g, self.mask)

            subscripts = f'{s[0:_g.ndim]},{s[0:alpha.ndim]}->{s[0:_g.ndim]}'

            if self.mask is None:
                resid = np.einsum(subscripts, _g, alpha) - self._f
                sse = 1/self._K * np.sum(np.square(resid))/np.sum(np.square(self.f))
            else:
                resid = self.mask * (np.einsum(subscripts, _g, alpha) - self._f)
                sse = 1/self._K * np.sum(np.square(resid))/np.sum(self.mask*np.square(self.f))

        else:
            alpha = np.array(1.0)
            if self.mask is None:
                resid = g - self.f
                numer = np.sum(np.square(resid), axis=(-2, -1))
                denom = np.sum(np.square(self.f), axis=(-2, -1))
                sse = 1/self._K * np.sum(numer/denom)
            else:
                resid = self.mask * (g - self.f)
                numer = np.sum(np.square(resid), axis=(-2, -1))
                denom = np.sum(self.mask * np.square(self.f), axis=(-2, -1))
                sse = 1/self._K * np.sum(numer/denom)

        self.cache_for_backward(resid, alpha)

        return float(sse)

    def backward(self, grad):
        # This function implements Eq. 53 in [2]
        resid, alpha = self.cache
        subs = f'{s[0:resid.ndim]},{s[0:alpha.ndim]}->{s[0:resid.ndim]}'

        if self.mask is None:
            grad = np.einsum(subs, resid, (2/self._D) * grad * alpha)
        else:
            grad = np.einsum(subs, self.mask * resid, (2/self._D) * grad * alpha)

        self.g.backward(grad)



import numpy as np

# This function was implemented in Numpy v1.12 and hasn't
# substantively changed since then (Numpy v1.20)
from numpy.core.einsumfunc import _parse_einsum_input

import loupe


# The implementation here is heavily influenced by chainer's einsum(). It
# has been simplified a bit and adapted to work with loupe's interfaces.

class einsum(loupe.core.Function):
    """
    einsum(subscripts, *operands, dtype=None)

    Evaluates the Einstein summation convention on the operands.

    Parameters
    ----------
    subscripts : str
        Specifies the subscripts for summation as comma separated list of subscript
        labels. An implicit (classical Einstein summation) calculation is performed 
        unless the explicit indicator ‘->’ is included as well as subscript labels 
        of the precise output form.
    operands : list of array_like
        The arrays used to perform the requested operation.

    Returns
    -------
    out : :class:`~loupe.core.Function`
        The result of the requested Einstein summation operation.
    
    See also
    --------
    :class:`~loupe.tensordot`

    """
    def __init__(self, *operands, dtype=None):
        # NOTE: I'm not sure this will work with complex inputs
        
        # Use numpy's _parse_einsum_input to tear apart the subscripts. Note
        # that although _parse_einsum_input also returns the operands, it 
        # casts them to numpy arrays which is not desirable here so we'll
        # just manually yank out the operands to preserve any loupe objects.
        in_subs, out_sub, _ = _parse_einsum_input(operands)
        opers = operands[1:] 

        self.in_subs = in_subs
        self.out_sub = out_sub
        self.operands = [loupe.asarray(oper) for oper in opers]
        
        super().__init__(*self.operands)
        
    @property
    def _inputs_requires_grad(self):
        # Return the indices of inputs with requires_grad = True
        return [n for n, s in enumerate(self.inputs) if s.requires_grad]
    
    def forward(self):
        subscripts = f'{self.in_subs}->{self.out_sub}'
        operands = [oper.getdata() for oper in self.operands]
        result = np.einsum(subscripts, *operands)
        
        self.cache_for_backward(*operands, result)
   
        return result
        
    def backward(self, grad):
        in_subs_fw = self.in_subs.split(',')
        out_sub_fw = self.out_sub
        
        *inputs, result = self.cache
        indices = self._inputs_requires_grad
        
        # compute gradient for inputs with requires_grad=True
        for i in indices:
            in_subs = ','.join([(out_sub_fw if j == i else s) for j, s in enumerate(in_subs_fw)])
            out_sub = in_subs_fw[i]
            out_shape = inputs[i].shape
            
            diag_inputs = [(grad if j == i else x) for j, x in enumerate(inputs)]
            out_set = set(out_sub)
            io_set = out_set.intersection(set(in_subs))
            
            if len(io_set) == len(out_sub):
                subscripts = f'{in_subs}->{out_sub}'
                y = np.einsum(subscripts, *diag_inputs)
            else:
                # I'm not sure what kind of expression will get us here ¯\_ (ツ)_/¯ 
                # See chainer DiagEinsum else block for implementation if it's needed
                raise NotImplementedError
                
            self.inputs[i].backward(y)




import numpy as np

import loupe

class add(loupe.core.Function):
    """Add arrays element-wise.

    Parameters
    ----------
    x1, x2 : array_like
        The arrays to be added. Arrays must either have the same shape or be
        broadcastable to a common shape (which becomes the shape of the
        output).

    Returns
    -------
    out : :class:`~loupe.core.Function`
        The sum of *x1* and *x2*, element-wise.

    Notes
    -----
    Equivalent to *x1 + x2*.

    Examples
    --------

    .. code:: pycon

        >>> loupe.add(1.0, 4.0)
        5.0
        >>> x1 = loupe.array([[1, 2], [3, 4]])
        >>> x2 = loupe.array([1, 2])
        >>> loupe.add(x1, x2)
        array([[ 2.,  4.],
               [ 4.,  6.]])
    
    The ``+`` operator can be used as shorthand for ``loupe.add``.

    .. code:: pycon

        >>> x1 = loupe.array([[1, 2], [3, 4]])
        >>> x2 = loupe.array([1, 2])
        >>> x1 + x2
        array([[ 2.,  4.],
               [ 4.,  6.]])

    """
    def __init__(self, x1, x2):
        self.left = loupe.asarray(x1)
        self.right = loupe.asarray(x2)
        super().__init__(self.left, self.right)

    def forward(self):
        left = self.left.getdata()
        right = self.right.getdata()
        self.cache_for_backward(left, right)
        
        result = np.add(left, right)
        return result
    
    def backward(self, grad):
        left, right = self.cache

        if self.left.requires_grad:
            left_grad = _broadcast_grad(grad, left.shape)
            self.left.backward(left_grad)

        if self.right.requires_grad:
            right_grad = _broadcast_grad(grad, right.shape)
            self.right.backward(right_grad)

class subtract(loupe.core.Function):
    """Subtract arrays element-wise.

    Parameters
    ----------
    x1, x2 : array_like
        The arrays to be subtracted from each other. Arrays must either have 
        the same shape or be broadcastable to a common shape (which becomes 
        the shape of the output).

    Returns
    -------
    out : :class:`~loupe.core.Function`
        The difference of *x1* and *x2*, element-wise.

    Notes
    -----
    Equivalent to *x1 - x2*.

    Examples
    --------

    .. code:: pycon

        >>> loupe.subtract(1.0, 4.0)
        -3.0
        >>> a = loupe.array([[1, 2], [3, 4]])
        >>> b = loupe.array([1, 2])
        >>> loupe.subtract(a, b)
        array([[ 0.,  0.],
               [ 2.,  2.]])
    
    The ``-`` operator can be used as shorthand for ``loupe.subtract``.

    .. code:: pycon

        >>> a = loupe.array([[1, 2], [3, 4]])
        >>> b = loupe.array([1, 2])
        >>> a - b
        array([[ 0.,  0.],
               [ 2.,  2.]])

    """
    def __init__(self, x1, x2):
        self.left = loupe.asarray(x1)
        self.right = loupe.asarray(x2)
        super().__init__(self.left, self.right)

    def forward(self):
        left = self.left.getdata()
        right = self.right.getdata()
        self.cache_for_backward(left, right)

        result = np.subtract(left, right)
        return result
    
    def backward(self, grad):
        
        left, right = self.cache

        if self.left.requires_grad:
            left_grad = _broadcast_grad(grad, left.shape)
            self.left.backward(left_grad)

        if self.right.requires_grad:
            right_grad = _broadcast_grad(-grad, right.shape)
            self.right.backward(right_grad) 

class multiply(loupe.core.Function):
    """Multiply arrays element-wise.

    Parameters
    ----------
    x1, x2 : array_like
        The arrays to be multiplied. Arrays must either have the same shape or 
        be broadcastable to a common shape (which becomes the shape of the
        output).

    Returns
    -------
    out : :class:`~loupe.core.Function`
        The product of *x1* and *x2*, element-wise.

    Notes
    -----
    Equivalent to *x1 * x2*.

    Examples
    --------

    .. code:: pycon

        >>> loupe.multiply(2.0, 4.0)
        8.0
        >>> x1 = loupe.array([[1, 2], [3, 4]])
        >>> x2 = loupe.array([1, 2])
        >>> loupe.multiply(x1, x2)
        array([[ 1.,  4.],
               [ 3.,  8.]])
    
    The ``*`` operator can be used as shorthand for ``loupe.multiply``.

    .. code:: pycon

        >>> x1 = loupe.array([[1, 2], [3, 4]])
        >>> x2 = loupe.array([1, 2])
        >>> x1 * x2
        array([[ 1.,  4.],
               [ 3.,  8.]])

    """
    def __init__(self, x1, x2):
        self.left = loupe.asarray(x1)
        self.right = loupe.asarray(x2)
        super().__init__(self.left, self.right)

    def forward(self):
        left = self.left.getdata()
        right = self.right.getdata()
        self.cache_for_backward(left, right)

        result = np.multiply(left, right)
        return result 

    def backward(self, grad):
    
        left, right = self.cache

        if self.left.requires_grad:
            left_grad = np.conj(right) * grad
            left_grad = _broadcast_grad(left_grad, left.shape)
            self.left.backward(left_grad)
        
        if self.right.requires_grad:
            right_grad = np.conj(left) * grad
            right_grad = _broadcast_grad(right_grad, right.shape)
            self.right.backward(right_grad)


def _broadcast_grad(grad, array_shape):
    # REF: https://github.com/joelgrus/autograd
    
    # sum out the added dimensions
    ndims_added = grad.ndim - len(array_shape)
    for _ in range(ndims_added):
        grad = grad.sum(axis=0)

    # sum across broadcasted (but not added) dimensions
    for n, dim in enumerate(array_shape):
        if dim == 1:
            grad = grad.sum(axis=n, keepdims=True)
    
    return grad


class power(loupe.core.Function):
    """First array elements raised to powers from second array, element-wise.

    Raise each base in *x1* to the positionally-corresponding power in *x2*. 
    *x1* and *x2* must be broadcastable to the same shape.

    Parameters
    ----------
    x1 : array_like
        Base array.
    x2 : array_like
        Exponent array.

    Returns
    -------
    out : :class:`~loupe.core.Function`
        The bases in *x1* raised to the exponent(s) in *x2*.

    Notes
    -----
    Equivalent to `x1 ** x2`.

    Examples
    --------
    Cube each element in an array.

    .. code:: pycon

        >>> x1 = loupe.array([1,2,3,4,5])
        >>> loupe.power(x1, 3)
        array([ 1., 8., 27., 64., 125.])

    Rase the bases to different exponents.

    .. code:: pycon

        >>> x2 = loupe.array([1,2,3,2,1])
        >>> loupe.power(x1, x2)
        array([ 1., 4., 27., 8., 5.]) 

    The `**` operator can be used as a shorthand for ``loupe.power``.

    .. code:: pycon

        >>> x = loupe.array([1,4,9,16])
        >>> x ** 0.5
        loupe.array([ 1., 2., 3., 4.])

    """
    def __init__(self, x1, x2):
        self.input = loupe.asarray(x1)
        self.exp = x2
        super().__init__(self.input)

    def forward(self):
        input = self.input.getdata()
        self.cache_for_backward(input)

        return np.power(input, self.exp)

    def backward(self, grad):
        input, = self.cache
        
        grad = self.exp * np.power(input, self.exp-1) * grad
        self.input.backward(grad)


class exp(loupe.core.Function):
    """Calculate the exponential of the input array.

    Parameters
    ----------
    x : array_like
        Input array

    Returns
    -------
    out : :class:`~loupe.core.Function`
        Output array, element-wise exponential of x.

    """
    def __init__(self, x):
        self.input = loupe.asarray(x)
        super().__init__(self.input)

    def forward(self):            
        result = np.exp(self.input.getdata())
        self.cache_for_backward(result)
        return result

    def backward(self, grad):
        result, = self.cache
        if np.any(np.iscomplex(result)):
            grad = grad * np.conj(result)
        else:
            grad *= result
        self.input.backward(grad)


class expc(loupe.core.Function):
    """Calculate the exponential of the input array * 1j.

    Parameters
    ----------
    x : array_like
        Input array

    Returns
    -------
    out : :class:`~loupe.core.Function`
        Output array, element-wise exponential of x*1j.

    Notes
    -----
    *loupe.expc(x)* is equivalent to *loupe.exp(x*1j)*.

    """
    def __init__(self, x):
        self.input = loupe.asarray(x)
        super().__init__(self.input)

    def forward(self):
        x = self.input.getdata()
        result = np.empty_like(x, dtype=complex)
        result.real = np.cos(x)
        result.imag = np.sin(x)
        self.cache_for_backward(result)
        return result

    def backward(self, grad):
        result, = self.cache
        grad = np.imag(grad * np.conj(result))
        # this is equivalent and maybe slightly faster?
        # grad = grad.imag * result.real - grad.real * result.imag
        self.input.backward(grad)


class sum(loupe.core.Function):
    """Sum of array elements.

    Parameters
    ----------
    x : array_like
        Array to be summed.

    Returns
    -------
    out : Function
        The sum of a.

    Note
    ----
    Currently sum only supports summing over all elements in the input array.

    Examples
    --------

    .. code:: pycon

        >>> loupe.sum([0.5, 1.5])
        2.0
        >>> loupe.sum([[1, 2], [3, 4]])
        10.0
    
    """

    # TODO: accept an axis parameter
    def __init__(self, x):
        self.input = loupe.asarray(x)
        super().__init__(self.input)

    def forward(self):
        result = self.input.getdata()
        self.cache_for_backward(result) 
        return np.sum(result)

    def backward(self, grad):
        result, = self.cache
        grad = np.full(result.shape, grad, dtype=result.dtype) 
        self.input.backward(grad)


class slice(loupe.core.Function):
    """Return a slice of the input array.

    Parameters
    ----------
    x : array_like
        Input array
    slc : slice
        Python slice object
        
    Returns
    -------
    out : :class:`~loupe.core.Function`
        Slice of *x*

    Notes
    -----
    :func:`array.__getitem__` is aliased to :class:`slice`, allowng slices to 
    be created using the familiar Python ``[]`` notation.

    """
    def __init__(self, x, slc):

        self.input = loupe.asarray(x)
        self.slc = slc
        super().__init__(self.input)

    def forward(self):
        result = self.input.getdata()
        self.cache_for_backward(result)
        return result[self.slc]

    def backward(self, grad):
        result, = self.cache
        result_grad = np.zeros(result.shape, dtype=result.dtype)
        result_grad[self.slc] = grad
        self.input.backward(result_grad)
        


from collections import namedtuple

import numpy as np
import scipy.optimize

import loupe

def optimize(fun, params, ftol=1e-9, gtol=1e-5, maxiter=1000,
             analytic_grad=True):
    """Minimize a scalar function of one or more variables using the
    L-BFGS-B algorithm.

    This function constructs a call to `scipy.optimize.minimize() 
    <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html>`_
    with `method = 'L-BFGS-B'`.

    Parameters
    ----------
    fun : :class:`~loupe.core.Function`
        The objective function to be minimized.
    params : list of :class:`array` or :class:`array`
        Parameters in `fun` to optimize.
    ftol : float, optional
        Termination tolerance on the objective function value. Default is 1e-9.
    gtol : float, optional
        Termination tolerance on the gradient value. Default is 1e-6.
    maxiter : int, optional
        Maximum number of iterations. Default is 1000.
    analytic_grad : bool, optional
        If true (default), the gradient vector is analytically computed using
        algorithmic differentiation. If false, the gradient is numerically
        estimated by a finite difference algorithm.

    Returns
    -------
    res : OptimizeResult
        The optimization result.  Important attributes are: ``x`` the solution 
        array, ``success`` a Boolean flag indicating if the optimizer exited 
        successfully and ``message`` which describes the cause of the 
        termination. See `OptimizeResult 
        <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.OptimizeResult.html>`_ 
        for a full description of other attributes.

    Examples
    --------
    Construct a simple quadratic objective function and compute the minimum
    value of x:

    .. code:: pycon

        >>> x = loupe.rand()
        >>> y = 4*(x-5)**2
        >>> loupe.optimize(y, params=x)

             fun: array(1.26217745e-29)
        hess_inv: <1x1 LbfgsInvHessProduct with dtype=float64>
             jac: array([-1.42108547e-14])
         message: 'CONVERGENCE: NORM_OF_PROJECTED_GRADIENT_<=_PGTOL'
            nfev: 3
             nit: 2
            njev: 3
          status: 0
         success: True
               x: array([5.])

    Optimize the same function, this time using the finite difference 
    gradient:

    .. code:: pycon

        >>> x = loupe.rand()
        >>> y = 4*(x-5)**2
        >>> loupe.optimize(y, params=x, analytic_grad=False)

             fun: array(8.4112054e-13)
        hess_inv: <1x1 LbfgsInvHessProduct with dtype=float64>
             jac: array([3.70850496e-06])
         message: 'CONVERGENCE: NORM_OF_PROJECTED_GRADIENT_<=_PGTOL'
            nfev: 6
             nit: 2
            njev: 3
          status: 0
         success: True
               x: array([5.00000046])
               
    """
    optim = Blueprint(fun, params)

    options = {'ftol': ftol, 'gtol': gtol, 'maxiter': maxiter}

    if analytic_grad:
        jac = optim.grad
    else: 
        jac = None

    res = scipy.optimize.minimize(optim.cost, optim.x, method='L-BFGS-B', 
                                  jac=jac, options=options)

    # NOTE: the solution returned by scipy.optimize.minimize is always one
    # step prior to the final step. (is this true for optimizers other than
    # L-BGFS-B or when the number of iterations constraint is hit?)

    # this means that optim.x is sometimes set in a much worse way than the
    # minimum solution. We'll upadte x here with the result
    optim.x = res.x

    return res


class Blueprint:
    def __init__(self, fun, params):

        params = [params] if not isinstance(params, (list, tuple)) else params
        _mark_params_for_optimization(fun, params)

        self.fun = fun
        self.params = []
        index = 0
        for p in params:
            size = p.flatten(apply_mask=True).size
            self.params.append((p, slice(index, index+size)))
            index += size

    @property
    def size(self):
        return self.params[-1][1].stop

    @property
    def x(self):
        x = np.empty(self.size, dtype=np.float64)
        for param in self.params:
            x[param[1]] = param[0].flatten(apply_mask=True)
        return x

    @x.setter
    def x(self, value):
        if value is not None:
            value = np.asarray(value)
            for param in self.params:
                if param[0].mask is not None:
                    # Optimizer only knows about valid values here. Need to
                    # dump these in the right spot in the full array
                    param[0][~param[0].mask] = value[param[1]]
                else:
                    param[0][...] = value[param[1]].reshape(param[0].shape) 


    def cost(self, x=None):
        # Evaluate the cost function at x
        if x is not None:
            if not np.all(x == self.x):
                self.x = x
        return self.fun.getdata()
    
    def grad(self, x=None):
        # Compute the gradient of all optimization parameters at x and 
        # return the result as a flattened array suitable for
        # consumption by an optimizer
        if x is not None: 
            if not np.all(x == self.x):
                self.x = x
        self.zero_grad()
        self.fun.backward(np.array(1.))
        
        grad = np.zeros(self.params[-1][1].stop, dtype=np.float64)
        for param in self.params:
            grad[param[1]] = param[0].grad_flatten(apply_mask=True)
        return grad

    def zero_grad(self):
        # Zero the gradients of all arrays in the graph
        for arr in _dump_arrays(self.fun):
            arr.zero_grad()


def _mark_params_for_optimization(fun, params):
    # Recursively trace through the graph and mark parameters to optimize
    # with requires_grad = True. All others should have 
    # requires_grad=False
    for arr in _dump_arrays(fun):
        if arr in params:
            arr.requires_grad = True
        else:
            arr.requires_grad = False  


def _dump_arrays(node, out=None):

    if out is None:
        out = []
    if isinstance(node, loupe.core.Function):
        for n in node.inputs:
            _dump_arrays(n, out)
    else:
        if node not in out:
            out.append(node)
    return out



import numpy as np

import loupe


class Node:
    """Base class for nodes in the computational graph"""

    # When return type is ambiguous to Numpy, prefer loupe Node objects over
    # Numpy ndarrays
    __array_priority__ = 1

    def __add__(self, other):
        return loupe.numeric.add(self, other)

    def __sub__(self, other):
        return loupe.numeric.subtract(self, other)

    def __rsub__(self, other):
        return loupe.numeric.subtract(other, self)

    def __mul__(self, other):
        return loupe.numeric.multiply(self, other)

    def __pow__(self, exponent):
        return loupe.numeric.power(self, exponent)

    __radd__ = __add__
    __rmul__ = __mul__

    def __array__(self):
        # REF: https://numpy.org/doc/stable/user/basics.dispatch.html
        return self.data
    
    def __str__(self):
        return np.array_str(self.data)

    def __repr__(self):
        return np.array_repr(self.data)

    def __getitem__(self, index):
        return loupe.numeric.slice(self, index)

    @property
    def data(self):
        """The array's data.

        Returns
        -------
        data : ndarray
        """
        pass

    @property
    def dtype(self):
        """Data type of the array's elements."""
        pass

    @property
    def shape(self):
        """Tuple of array dimensions."""
        pass

    @property
    def requires_grad(self):
        pass

    def getdata(self, copy=False, dtype=None):
        """Return the array's data.

        Parameters
        ----------
        copy : bool, optional
            Whether to force a copy of the underlying data to be returned.
        dtype : type or numpy dtype, optional
            The dtype of the returned data.

        """
        data = self.data.copy() if copy else self.data
        if dtype is None:
            dtype = self.dtype
        return data.astype(dtype)


class array(Node):
    """Create a multidimensional array.
    
    Parameters
    ----------
    object : array_like
        An array or other data type that can be interpreted as an array.
    dtype : data-type, optional
        The desired data-type for the array. If not given, the type will be
        inferred from the supplied data object.
    mask : array_like, optional
        Mask applied to the array where a True value indicates that the 
        corresponding element of the array is invalid. Mask must have the same 
        shape as the array and contain entries that are castable to bool. If 
        None (default), the array is not masked.
    requires_grad : bool, optional
        It True, gradients will be computed for this array. Default is False.
    
    """

    # TODO: masking, scaling, bounds?

    def __init__(self, object, dtype=None, mask=None, requires_grad=False):

        self._data = np.asarray(object, dtype=dtype)
        self.mask = mask
        self.requires_grad = requires_grad
        self.grad = np.zeros(self._data.shape, dtype=float)

    def __setitem__(self, index, value):
        value = np.asarray(value, dtype=self.dtype)
        self._data[index] = value

    @property
    def data(self):
        return self._data

    @property
    def dtype(self):
        return self.data.dtype

    @property
    def shape(self):
        return self.data.shape
    
    @property
    def ndim(self):
        return self.data.ndim
    
    @property
    def strides(self):
        return self.data.strides

    @property
    def mask(self):
        return self._mask
    
    @mask.setter
    def mask(self, value):
        if value is None:
            self._mask = None
        else:
            mask = np.asarray(value, dtype=bool)
            if mask.shape != self.shape:
                raise ValueError('Mask must have same shape as array')
            self._mask = mask

    @property
    def requires_grad(self):
        """Is ``True`` if gradients need to be computed for this 
        :class:`array`, ``False`` otherwise.
        """
        return self._requires_grad

    @requires_grad.setter
    def requires_grad(self, value):
        self._requires_grad = value

    @property
    def grad(self):
        """Gradient of the array. 
        
        This attribute is zero by default and is only computed when 
        :attr:`requires_grad` is True and 
        :func:`~loupe.Function.backward` is called.

        Returns
        -------
        grad : ndarray

        """
        return self._grad

    @grad.setter
    def grad(self, value):
        self._grad = value

    def zero_grad(self):
        """Zero the array gradient."""
        self.grad = np.zeros_like(self._data)

    def flatten(self, apply_mask=False):
        """Return the array flattened into one dimension.
        
        Parameters
        ----------
        apply_mask : bool, optional
            If True, only non-masked data is returned. Default is False.
        """
        if self.mask is not None and apply_mask:
            # Since True corresponds to data to be masked, ~mask is the slice
            # we really want. Note this operation implicitly flattens the data
            return self._data[~self.mask]
        else:
            return self._data.flatten()

    def grad_flatten(self, apply_mask=False):
        if self.mask is not None and apply_mask:
            return self._grad[~self.mask]
        else:
            return self._grad.flatten()

    def backward(self, grad):
        """Compute the gradient of the array.

        Parameters
        ----------
        grad : array_like
            Gradient with respect to the array.

        """
        if not self.requires_grad:
            pass
        else:
            # accumulate gradient
            if self.dtype in (np.float32, np.float64):
                self.grad += grad.real
            else:
                self.grad += grad


class Function(Node):
    """Base class for representing functions that operate on :class:`~loupe.array` 
    objects.

    `Function` defines the interface for objects that operate on 
    :class:`~loupe.array` objects but it doesn't actually implement any useful
    functionality. See :ref:`extend` for more details on how to define a new
    operation by subclassing `Function`.

    Parameters
    ----------
    inputs : list of :class:`~loupe.array`
        Array objects that the function operates on.

    See also
    --------
     A complete list of available functions is available
    :ref:`here <api.functions>`.

    """
    def __init__(self, *inputs):
        self._inputs = inputs
        self._cache = []
        self._data = self.forward()

    @property
    def data(self):
        self._data = np.asarray(self.forward())
        return self._data

    @property
    def dtype(self):

        return self._data.dtype

    @property
    def shape(self):
        return self._data.shape
    
    @property
    def ndim(self):
        return self._data.ndim

    @property
    def inputs(self):
        """List of array inputs the function operates on."""
        return self._inputs

    @property
    def requires_grad(self):
        """Is ``True`` if any of the function's inputs have 
        ``requires_grad=True``, ``False`` otherwise.
        """
        return any([input.requires_grad for input in self.inputs])

    @property
    def cache(self):
        """Data that has been saved by :func:`cache_for_backward`.

        Returns
        -------
        cache : list

        """
        return self._cache

    def cache_for_backward(self, *data):
        """Save data for a future call to :func:`backward`.

        Saved data can be accessed through the :attr:`cache` attribute. 

        Parameters
        ----------
        data : list_like
            List of data to cache. Cached data should *probably* be ndarrays,
            but nothing prevents the caching of any Python type.

        Warning
        -------
        Care must be taken that cached data is not modified in place after
        calling :func:`forward`. Doing so will make the cached data stale and
        its resulting use during a call to :func:`backward` will be invalid.

        Notes
        -----
        This method should be called at most once, and only from inside the 
        :func:`forward` method.

        """
        self._cache = data

    def forward(self):
        """Performs the operation.

        This method should by overridden by all subclasses.

        Returns
        -------
        out : ndarray
            The result of the operation.

        """
        raise NotImplementedError
        
    def backward(self, grad):
        """Defines a formula for differentiating the operation.

        This method should be overridden by all subclasses.

        Parameters
        ----------
        grad : array_like
            The gradient with respect to the function output.

        """
        raise NotImplementedError



import numpy as np
from numpy.lib.stride_tricks import as_strided

import loupe


class absolute_square(loupe.core.Function):
    """Calculate the absolute value squared, element-wise.

    Parameters
    ----------
    x : array_like
        Input array.

    Returns
    -------
    out : :class:`~loupe.core.Function`
        Element-wise *abs(x**2)*

    Notes
    -----
    :class:`abs_square` is an alias for :class:`absolute_square`.
    
    """
    def __init__(self, x):
        self.input = loupe.asarray(x)
        super().__init__(self.input)

    def forward(self):
        input = self.input.getdata()
        result = np.square(np.abs(input))
        self.cache_for_backward(input)
        return result

    def backward(self, grad):
        input, = self.cache
        grad = 2 * input * grad
        self.input.backward(grad)


abs_square = absolute_square


class rebin(loupe.core.Function):
    """Rebin an array.  

    Parameters
    ----------
    x : array_like
        Input array
    bin_shape: tuple
        The shape of the bin. Each dimension must divide evenly into the
        corresponsing dimension of `x`.
    
    Returns
    -------
    out : :class:`~loupe.core.Function`
        Rebinned array
    """

    def __init__(self, x, bin_shape):
        self.input = loupe.asarray(x)
        self.bin_shape =  np.broadcast_to(bin_shape, (2,))
        
        if (self.bin_shape <= 0).any():
            raise ValueError("'bin_shape' elements must be positive")
        
        if (self.input.shape % self.bin_shape).sum() != 0:
            raise ValueError("'bin_shape' is not compatible with input array shape")

        super().__init__(self.input)

    def forward(self):
        input = self.input.getdata()

        # the code below is adapted from the skimage function view_as_blocks()
        new_shape = tuple(input.shape // self.bin_shape) + tuple(self.bin_shape)
        new_strides = tuple(input.strides * self.bin_shape) + input.strides
        restrided_input = as_strided(input, shape=new_shape, strides=new_strides)

        result = np.einsum('ijkl->ij', restrided_input)
        
        self.cache_for_backward(input.shape, result.shape)

        return result

    def backward(self, grad):
        input_shape, result_shape, = self.cache

        if result_shape != grad.shape:
            raise ValueError(f'grad must have shape = {result_shape}')
        
        new_shape = tuple(grad.shape) + tuple(self.bin_shape)
        new_strides = tuple(grad.strides) + (0, 0)
        restrided_grad = as_strided(grad, shape=new_shape, strides=new_strides)
        grad = np.copy(np.swapaxes(restrided_grad, 1, 2)).reshape(input_shape)

        self.input.backward(grad)


class normavg(loupe.core.Function):

    def __init__(self, x):
        self.input = loupe.asarray(x)
        super().__init__(self.input)

    def forward(self):
        input = self.input.getdata()
        mean = np.sum(input)
        result = np.reciprocal(mean) * input
        self.cache_for_backward(mean)
        return result

    def backward(self, grad):
        print('grad pre:', np.mean(grad))
        mean, = self.cache
        grad = np.reciprocal(mean) * grad
        #grad += -1*np.reciprocal(mean)
        print('grad_post:', np.mean(grad))
        self.input.backward(grad)





import numpy as np


def mesh(shape, shift=(0, 0)):
    """Generate a standard mesh."""

    nr = shape[0]
    nc = shape[1]

    x = np.arange(shape[1]) - np.floor(shape[1]/2.0) - shift[1]
    y = np.arange(shape[0]) - np.floor(shape[0]/2.0) - shift[0]

    return np.meshgrid(x, y)


def circle(shape, radius, shift=(0, 0)):
    """Compute a circle with anti-aliasing.

    Parameters
    ----------
    shape : array_like
        Size of output in pixels (nrows, ncols)

    radius : float
        Radius of circle in pixels

    shift : (2,) array_like, optional
        How far to shift center in float (rows, cols). Default is (0, 0).

    Returns
    -------
    out : ndarray

    """
    x, y = mesh(shape)
    r = np.sqrt(np.square(x - shift[1]) + np.square(y - shift[0]))
    return np.clip(radius + 0.5 - r, 0.0, 1.0)


def circlemask(shape, radius, shift=(0, 0)):
    """Compute a circular mask.

    Parameters
    ----------
    shape : array_like
        Size of output in pixels (nrows, ncols)

    radius : float
        Radius of circle in pixels

    shift : array_like, optional
        How far to shift center in float (rows, cols). Default is (0, 0).

    Returns
    -------
    out : ndarray

    """

    mask = circle(shape, radius, shift)
    mask[mask > 0] = 1
    return mask


def hexagon(shape, radius, rotate=False):
    """Compute a hexagon mask.

    Parameters
    ----------
    shape : array_like
        Size of output in pixels (nrows, ncols)

    radius : int
        Radius of outscribing circle (which also equals the side length) in
        pixels.

    rotate : bool
        Rotate mask so that flat sides are aligned with the Y direction instead
        of the default orientation which is aligned with the X direction.

    Returns
    -------
    out : ndarray

    """

    inner_radius = radius * np.sqrt(3)/2
    side_length = radius/2

    y, x = mesh(shape)

    rect = np.where((np.abs(x) <= side_length) & (np.abs(y) <= inner_radius))
    left_tri = np.where((x <= -side_length) & (x >= -radius) & (np.abs(y) <= (x + radius)*np.sqrt(3)))
    right_tri = np.where((x >= side_length) & (x <= radius) & (np.abs(y) <= (radius - x)*np.sqrt(3)))

    mask = np.zeros(shape)
    mask[rect] = 1
    mask[left_tri] = 1
    mask[right_tri] = 1

    if rotate:
        return mask.transpose()
    else:
        return mask


def slit(shape, width):
    """Compute a horizontal slit mask.

    Parameters
    ----------
    shape : array_like
        Size of output in pixels (nrows, ncols)

    width : int
        Slit width in pixels

    Returns
    -------
    out : ndarray

    """

    y, x = mesh(shape)

    mask = np.zeros(shape)
    mask[np.abs(x) <= width/2] = 1

    return mask


import numpy as np
from numpy.lib.stride_tricks import as_strided

import loupe

def asarray(a):
    """Convert the input to an array.
    
    Parameters
    ----------
    a : array_like
        Input data, in any form that can be converted to an array. This 
        includes lists, lists of tuples, tuples, tuples of tuples, 
        tuples of lists and ndarrays.

    Returns
    -------
    out : :class:`array`
        Input data packaged as an :class:`array`. If input is already an
        :class:`array` or :class:`~loupe.core.Function`, the original input is
        returned.
    
    Examples
    --------
    Convert a list into an array:

    .. code:: pycon

        >>> a = [1,2,3]
        >>> loupe.asarray(a)
        array([ 1.,2., 3.])

    Existing arrays are not copied:

    .. code:: pycon

        >>> a = loupe.array([1,2,3])
        >>> loupe.asarray(a) is a
        True

    """
    if isinstance(a, (loupe.array, loupe.core.Function)):
        return a
    elif isinstance(a, (int, float, complex, list, tuple, np.ndarray)):
        return loupe.array(a)
    else:
        raise TypeError(f'Unsupported type. Cannot create array from \
                          {a.__class__.__name__}')


def zeros(shape, dtype=None, requires_grad=False):
    """Return a new array with values set to zeros.
    
    Parameters
    ----------
    shape : int or tuple of ints
        Shape of the array
    dtype : data type, optional
        Desired data type for the array
    requires_grad : bool, optional
        It True, gradients will be computed for this array. Default is False.

    Returns
    -------
    out : :class:`array`
        Array of zeros with the given shape and dtype.

    See Also 
    --------
    zeros_like : Return an array of zeros with shape and type of input.
    ones : Return a new array with values set to ones.
    
    """
    return loupe.array(np.zeros(shape), dtype=dtype, 
                       requires_grad=requires_grad)


def zeros_like(a, dtype=None, requires_grad=False):
    """Return a new array of zeros with the same shape and type as a given 
    array.
    
    """
    return loupe.zeros(shape=a.shape, dtype=dtype, 
                       requires_grad=requires_grad)

def ones(shape, dtype=None, requires_grad=False):
    """Return a new array with values set to ones."""
    return loupe.array(np.ones(shape), dtype=dtype, 
                       requires_grad=requires_grad)


def ones_like(a, dtype=None, requires_grad=False):
    """Return a new array of ones with the same shape and type as a given 
    array.
    
    """
    return loupe.ones(shape=a.shape, dtype=dtype, 
                      requires_grad=requires_grad)


def rand(low=0.0, high=1.0, size=None, dtype=None, requires_grad=False):
    """Return a new array with values drawn from a uniform distribution."""
    return loupe.array(np.random.uniform(low=low, high=high, size=size), 
                 dtype=dtype, requires_grad=requires_grad)


def randn(loc=0.0, std=1.0, size=None, dtype=None, requires_grad=False):
    """Return a new array with values drawn from a normal distribution."""
    return loupe.array(np.random.normal(loc=loc, scale=std, size=size), dtype=dtype, 
                       requires_grad=requires_grad)


def centroid(img):
    """Compute image centroid location.

    Parameters
    ----------
    img : array_like
        Input array.

    Returns
    -------
    centroid : tuple
        ``(x,y)`` centroid location.

    """
    img = np.asarray(img)
    img = img/np.sum(img)
    nr, nc = img.shape
    yy, xx = np.mgrid[0:nr, 0:nc]

    x = np.dot(xx.ravel(), img.ravel())
    y = np.dot(yy.ravel(), img.ravel())

    return x, y

def shift(a, shift):
    """Shift an array via FFT.

    Shift an array by (row, column). The shifts may be non-integer as the 
    shift operation is implemented by introducing a Fourier-domain tilt. If 
    ``a`` is complex, the result will also be complex.

    Parameters
    ----------
    a : array_like
        The input array.
    shift : (2,) sequence
        The shift specified as (row, column).
    Returns
    -------
    shifted : ndarray
        The shifted input array.

    Example
    -------
    .. code:: pycon

        >>> import loupe
        >>> import numpy as np
        >>> arr = np.zeros((3,3))
        >>> arr[2,2] = 1
        >>> arr
        array([[0., 0., 0.],
               [0., 0., 0.],
               [0., 0., 1.]])
        >>> arr_shift = loupe.shift(arr, shift=(-1,-1))
        >>> arr_shift
        array([[ 0.00000000e+00, -7.40148683e-17, -2.46716228e-17],
               [-1.16747372e-16,  1.00000000e+00,  2.14548192e-16],
               [-3.12823642e-17,  2.22044605e-16, -4.18468327e-17]])
    """
    a = np.asarray(a)

    dr, dc = shift
    R = dr * np.fft.fftfreq(a.shape[0])
    C = dc * np.fft.fftfreq(a.shape[1])

    RR, CC = np.meshgrid(R, C, indexing='ij')
    K = np.exp(-1j*2*np.pi*(RR+CC))
    shifted = np.fft.ifft2(np.fft.fft2(a)*K)

    if np.any(np.iscomplex(a)):
        return shifted
    else:
        return shifted.real


def register(arr, ref, oversample, return_error=False):
    """Compute the subpixel image translation to register the input array to a
    reference array.

    The registration shift is computed in two steps: first a coarse estimate 
    is computed from the FFT-based cross-correlation of the two input arrays.
    This estimate is then refined to subpixel accuracy by computing the 
    upsampled DFT-based cross-correlation in a small neigborhood around the
    initial estimate.

    Parameters
    ----------
    arr : array_like
        Array to register.
    ref : array_like
        Target array.
    oversample : float
        Oversampling factor for subpixel registration. Registration accuracy
        is 1/oversample.
    return_error : bool, optional
        If True, the noramlized RMS registration error is returned. Default is
        False.
    Returns
    -------
    shift : tuple
        Translation in (row, col) that will register *arr* to *ref*.
    err : float
        Registration error

    References
    ----------
    Guizar-Sicairos, Thurman, and Fienup, "Efficient subpixel image 
    registration algorithms". Optics Letters 33, 156-158 (2008)

    See also
    --------
    :func:`~loupe.shift`

    Example
    -------
    .. code:: pycon

        >>> import loupe
        >>> import numpy as np
        >>> ref = np.zeros((3,3))
        >>> ref[1,1] = 1
        >>> arr = np.zeros((3,3))
        >>> arr[2,2] = 1
        >>> shift = loupe.register(arr, ref, oversample=2)
        >>> shift
        (-1.0, -1.0)

    """
    F = np.fft.fft2(arr)
    G = np.fft.fft2(ref)
    xcorr = np.fft.fftshift(np.fft.ifft2(G*np.conj(F)))
    
    # find peak
    maxima = np.unravel_index(np.argmax(np.abs(xcorr)), xcorr.shape)
    peak = xcorr[maxima]

    # compute shifts
    center = np.array([np.fix(x/2) for x in arr.shape])
    shift = maxima - center
    if oversample !=1:
        # now we can set up and perform the oversampled dft on an oversampled 
        # 1.5 x 1.5 pixel region about the peak
        npix_dft = np.ceil(oversample*1.5)
        dft_shift = np.fix(npix_dft/2)
        rs = dft_shift - shift[0] * oversample
        cs = dft_shift - shift[1] * oversample

        # Compute DFT
        X = np.arange(arr.shape[1]) - np.floor(arr.shape[1]/2)
        Y = np.arange(arr.shape[0]) - np.floor(arr.shape[0]/2)
        U = np.arange(npix_dft) - cs
        V = np.arange(npix_dft) - rs
        E1 = np.exp(-2*np.pi*1j/(arr.shape[0]*oversample)*np.outer(V,Y))
        E2 = np.exp(-2*np.pi*1j/(arr.shape[1]*oversample)*np.outer(X,U))
        xcorr = np.dot(np.dot(E1,np.conj(np.fft.ifftshift(G*np.conj(F)))),E2)
        
        maxima_subpx = np.unravel_index(np.argmax(np.abs(xcorr)), xcorr.shape)
        peak = xcorr[maxima_subpx]

        # Combine subpixel peak coordinates with integer pixel peak coords
        maxima_subpx -= dft_shift 
        shift[0] += maxima_subpx[0]/oversample
        shift[1] += maxima_subpx[1]/oversample

    shift = tuple(shift)

    # Compute normalized RMS error
    if return_error:
        arr_amp = np.sum(np.abs(F)**2)
        ref_amp = np.sum(np.abs(G)**2)
        err = 1-np.abs(peak)**2/(arr_amp*ref_amp)
        err = np.sqrt(np.abs(err))
        return shift, err

    return shift

def medfix2(input, mask, kernel=(3,3)):
    """Fix masked entries in a 2-dimensional array via median filtering.

    Parameters
    ----------
    input : array_like
        A 2-dimensional input array
    mask : array_like
        A 2-dimensional
    kernel : array_like, optional
        A scalar or list of length 2 specifying the filter window in 
        each dimension. Elements of *kernel* should be odd. If *kernel*
        is a scalar, it is used for each dimension. Default is (3,3).
    Returns
    -------
    out : ndarray
        An array the same size as input where the masked entries are 
        replaced with median filtered values.
    
    """
    # force a copy by calling array instead of asarray
    input = np.array(input, dtype=float)

    mask = np.asarray(mask, dtype=bool)

    kernel = np.asarray(kernel)
    if kernel.shape == ():
        kernel = np.repeat(kernel, 2)
    if np.any(kernel % 2 == 0):
        raise ValueError("Each element of kernel must be odd")

    input[mask] = np.nan

    pad_size = np.array((kernel-1)/2, dtype=int)
    input_pad = np.full((input.shape[0] + pad_size[0] * 2, 
                         input.shape[1] + pad_size[1] * 2), np.nan)
    slc = (slice(pad_size[0], pad_size[0] + input.shape[0]), 
           slice(pad_size[1], pad_size[1] + input.shape[1]))
    input_pad[slc] = input

    input_pad = as_strided(input_pad, 
                           shape=(input_pad.shape[0], input_pad.shape[1], kernel[0], kernel[1]),
                           strides=input_pad.strides + input_pad.strides)
    input_pad = input_pad.reshape((input_pad.shape[0], input_pad.shape[1], np.prod(kernel)))

    mask_idx = np.where(mask)

    input[mask_idx] = np.nanmedian(input_pad[mask_idx[0], mask_idx[1], :], axis=1)

    return input
